from typing import Annotated

import nshconfig as C

from .base import CallbackConfigBase as CallbackConfigBase
from .early_stopping import EarlyStopping as EarlyStopping
from .ema import EMA as EMA
from .ema import EMAConfig as EMAConfig
from .finite_checks import FiniteChecksCallback as FiniteChecksCallback
from .finite_checks import FiniteChecksConfig as FiniteChecksConfig
from .gradient_skipping import GradientSkipping as GradientSkipping
from .gradient_skipping import GradientSkippingConfig as GradientSkippingConfig
from .interval import EpochIntervalCallback as EpochIntervalCallback
from .interval import IntervalCallback as IntervalCallback
from .interval import StepIntervalCallback as StepIntervalCallback
from .latest_epoch_checkpoint import LatestEpochCheckpoint as LatestEpochCheckpoint
from .latest_epoch_checkpoint import (
    LatestEpochCheckpointCallbackConfig as LatestEpochCheckpointCallbackConfig,
)
from .log_epoch import LogEpochCallback as LogEpochCallback
from .model_checkpoint import ModelCheckpoint as ModelCheckpoint
from .model_checkpoint import (
    ModelCheckpointCallbackConfig as ModelCheckpointCallbackConfig,
)
from .norm_logging import NormLoggingCallback as NormLoggingCallback
from .norm_logging import NormLoggingConfig as NormLoggingConfig
from .on_exception_checkpoint import OnExceptionCheckpoint as OnExceptionCheckpoint
from .on_exception_checkpoint import (
    OnExceptionCheckpointCallbackConfig as OnExceptionCheckpointCallbackConfig,
)
from .print_table import PrintTableMetricsCallback as PrintTableMetricsCallback
from .print_table import PrintTableMetricsConfig as PrintTableMetricsConfig
from .throughput_monitor import ThroughputMonitorConfig as ThroughputMonitorConfig
from .timer import EpochTimer as EpochTimer
from .timer import EpochTimerConfig as EpochTimerConfig
from .wandb_watch import WandbWatchCallback as WandbWatchCallback
from .wandb_watch import WandbWatchConfig as WandbWatchConfig

CallbackConfig = Annotated[
    ThroughputMonitorConfig
    | EpochTimerConfig
    | PrintTableMetricsConfig
    | FiniteChecksConfig
    | NormLoggingConfig
    | GradientSkippingConfig
    | EMAConfig
    | ModelCheckpointCallbackConfig
    | LatestEpochCheckpointCallbackConfig
    | OnExceptionCheckpointCallbackConfig
    | WandbWatchConfig,
    C.Field(discriminator="name"),
]



import logging
from pathlib import Path
from typing import Literal

from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks import Checkpoint
from typing_extensions import override

from .base import CallbackConfigBase

log = logging.getLogger(__name__)


class LatestEpochCheckpointCallbackConfig(CallbackConfigBase):
    kind: Literal["latest_epoch_checkpoint"] = "latest_epoch_checkpoint"

    dirpath: str | Path | None = None
    """Directory path to save the checkpoint file."""

    filename: str = "latest_epoch{epoch:02d}_step{step:04d}.ckpt"
    """Checkpoint filename. This must not include the extension."""

    save_weights_only: bool = False
    """Whether to save only the model's weights or the entire model object."""

    latest_symlink_filename: str | None = "latest.ckpt"
    """Filename for the latest symlink. If None, no symlink will be created."""

    @override
    def create_callbacks(self, root_config):
        dirpath = self.dirpath or root_config.directory.resolve_subdirectory(
            root_config.id, "checkpoint"
        )
        dirpath = Path(dirpath)

        yield LatestEpochCheckpoint(self, dirpath)


class LatestEpochCheckpoint(Checkpoint):
    def __init__(self, config: LatestEpochCheckpointCallbackConfig, dirpath: Path):
        super().__init__()

        self.config = config
        self.dirpath = dirpath

        # Also, we hold a reference to the last checkpoint path
        # to be able to remove it when a new checkpoint is saved.
        self._last_ckpt_path: Path | None = None

    def _ckpt_path(self, trainer: Trainer):
        return self.dirpath / self.config.filename.format(
            epoch=trainer.current_epoch, step=trainer.global_step
        )

    @override
    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule):
        # Remove the last checkpoint if it exists
        if self._last_ckpt_path is not None:
            trainer.strategy.remove_checkpoint(self._last_ckpt_path)

        # Save the new checkpoint
        filepath = self._ckpt_path(trainer)
        trainer.save_checkpoint(filepath, self.config.save_weights_only)
        self._last_ckpt_path = filepath

        # Create the latest symlink
        if (symlink_filename := self.config.latest_symlink_filename) is not None:
            symlink_path = self.dirpath / symlink_filename
            if symlink_path.exists():
                symlink_path.unlink()
            symlink_path.symlink_to(filepath.name)
            log.info(f"Created latest symlink: {symlink_path}")

    def latest_checkpoint(self):
        if (symlink_filename := self.config.latest_symlink_filename) is None:
            return None

        if not (symlink_path := self.dirpath / symlink_filename).exists():
            return None

        return symlink_path



import copy
from collections.abc import Callable
from typing import Any, cast

from typing_extensions import TypeVar

TDataset = TypeVar("TDataset", infer_variance=True)


def transform(
    dataset: TDataset,
    transform: Callable[[Any], Any],
    *,
    deepcopy: bool = False,
) -> TDataset:
    """
    Wraps a dataset with a transform function.

    Args:
        dataset: The dataset to wrap.
        transform: The transform function to apply to each item.
        deepcopy: Whether to deep copy each item before applying the transform.
    """

    import wrapt

    class _TransformedDataset(wrapt.ObjectProxy):
        def __getitem__(self, idx):
            nonlocal deepcopy, transform

            data = self.__wrapped__.__getitem__(idx)
            if deepcopy:
                data = copy.deepcopy(data)
            data = transform(data)
            return data

    return cast(TDataset, _TransformedDataset(dataset))


def transform_with_index(
    dataset: TDataset,
    transform: Callable[[Any, int], Any],
    *,
    deepcopy: bool = False,
) -> TDataset:
    """
    Wraps a dataset with a transform function that takes an index, in addition to the item.

    Args:
        dataset: The dataset to wrap.
        transform: The transform function to apply to each item.
        deepcopy: Whether to deep copy each item before applying the transform.
    """

    import wrapt

    class _TransformedWithIndexDataset(wrapt.ObjectProxy):
        def __getitem__(self, idx: int):
            nonlocal deepcopy, transform

            data = self.__wrapped__.__getitem__(idx)
            if deepcopy:
                data = copy.deepcopy(data)
            data = transform(data, idx)
            return data

    return cast(TDataset, _TransformedWithIndexDataset(dataset))



import contextlib
from pathlib import Path
from typing import Literal

from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks.callback import Callback
from typing_extensions import TypeAlias, override

from .base import CallbackConfigBase

try:
    from nshutils import ActSave  # type: ignore
except ImportError:
    ActSave = None

Stage: TypeAlias = Literal["train", "validation", "test", "predict"]


class ActSaveConfig(CallbackConfigBase):
    enabled: bool = True
    """Enable activation saving."""

    save_dir: Path | None = None
    """Directory to save activations to. If None, will use the activation directory set in `config.directory`."""

    def __bool__(self):
        return self.enabled

    @override
    def create_callbacks(self, root_config):
        yield ActSaveCallback(
            self,
            self.save_dir
            or root_config.directory.resolve_subdirectory(root_config.id, "activation"),
        )


class ActSaveCallback(Callback):
    def __init__(self, config: ActSaveConfig, save_dir: Path):
        super().__init__()

        self.config = config
        self.save_dir = save_dir
        self._enabled_context: contextlib._GeneratorContextManager | None = None
        self._active_contexts: dict[Stage, contextlib._GeneratorContextManager] = {}

    @override
    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        super().setup(trainer, pl_module, stage)

        if not self.config:
            return

        if ActSave is None:
            raise ImportError(
                "ActSave is not installed. Please install nshutils to use the ActSaveCallback."
            )

        context = ActSave.enabled(self.save_dir)
        context.__enter__()
        self._enabled_context = context

    @override
    def teardown(
        self, trainer: Trainer, pl_module: LightningModule, stage: str
    ) -> None:
        super().teardown(trainer, pl_module, stage)

        if not self.config:
            return

        if self._enabled_context is not None:
            self._enabled_context.__exit__(None, None, None)
            self._enabled_context = None

    def _on_start(self, stage: Stage, trainer: Trainer, pl_module: LightningModule):
        if not self.config:
            return

        if ActSave is None:
            raise ImportError(
                "ActSave is not installed. Please install nshutils to use the ActSaveCallback."
            )

        # If we have an active context manager for this stage, exit it
        if active_contexts := self._active_contexts.get(stage):
            active_contexts.__exit__(None, None, None)

        # Enter a new context manager for this stage
        context = ActSave.context(stage)
        context.__enter__()
        self._active_contexts[stage] = context

    def _on_end(self, stage: Stage, trainer: Trainer, pl_module: LightningModule):
        if not self.config:
            return

        # If we have an active context manager for this stage, exit it
        if (active_contexts := self._active_contexts.pop(stage, None)) is not None:
            active_contexts.__exit__(None, None, None)

    @override
    def on_train_epoch_start(self, trainer: Trainer, pl_module: LightningModule):
        return self._on_start("train", trainer, pl_module)

    @override
    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule):
        return self._on_end("train", trainer, pl_module)

    @override
    def on_validation_epoch_start(self, trainer: Trainer, pl_module: LightningModule):
        return self._on_start("validation", trainer, pl_module)

    @override
    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule):
        return self._on_end("validation", trainer, pl_module)

    @override
    def on_test_epoch_start(self, trainer: Trainer, pl_module: LightningModule):
        return self._on_start("test", trainer, pl_module)

    @override
    def on_test_epoch_end(self, trainer: Trainer, pl_module: LightningModule):
        return self._on_end("test", trainer, pl_module)

    @override
    def on_predict_epoch_start(self, trainer: Trainer, pl_module: LightningModule):
        return self._on_start("predict", trainer, pl_module)

    @override
    def on_predict_epoch_end(self, trainer: Trainer, pl_module: LightningModule):
        return self._on_end("predict", trainer, pl_module)



import weakref
from typing import Set

import torch
from torch.autograd.graph import register_multi_grad_hook
from torch.nn.modules.module import (
    register_module_forward_hook,
    register_module_forward_pre_hook,
)
from torch.utils._pytree import tree_flatten

__all__ = ["ModuleTracker"]


class ModuleTracker:
    """
    ``ModuleTracker`` is a context manager that tracks the nn.Module hierarchy during execution
    so that other system can query which Module is currently being executed (or its backward is being
    executed).

    You can access the ``parents`` attribute on this context manager to get the set of all the
    Modules currently being executed via their fqn (fully qualified name, also used as the key within
    the state_dict).
    You can access the ``is_bw`` attribute to know if you are currently running in backward or not.

    Note that ``parents`` is never empty and always contains the "Global" key. The ``is_bw`` flag
    will remain ``True`` after the forward until another Module is executed. If you need it to be
    more accurate, please submit an issue requesting this. Adding a map from fqn to the module instance
    is possible but not done yet, please submit an issue requesting this if you need it.

    Example usage

    .. code-block:: python

        mod = torch.nn.Linear(2, 2)

        with ModuleTracker() as tracker:
            # Access anything during the forward pass
            def my_linear(m1, m2, bias):
                print(f"Current modules: {tracker.parents}")
                return torch.mm(m1, m2.t()) + bias
            torch.nn.functional.linear = my_linear

            mod(torch.rand(2, 2))

    """

    parents: Set[str]
    """
    A Set containing the fqn for each module currently running their forward
    """

    def __init__(self):
        self.parents = {"Global"}
        self._known_modules: weakref.WeakKeyDictionary = weakref.WeakKeyDictionary()
        self._seen_modules: weakref.WeakSet = weakref.WeakSet()
        self._has_callback = False

    def _maybe_set_engine_callback(self):
        # This assumes no concurrent calls to backward
        if self._has_callback:
            return

        def callback():
            self.parents = {"Global"}
            self._has_callback = False

        torch.autograd.Variable._execution_engine.queue_callback(callback)
        self._has_callback = True

    @property
    def is_bw(self):
        """
        A boolean marking if this is currently running during the backward pass or not
        """
        return torch._C._current_graph_task_id() != -1

    def _get_mod_name(self, mod):
        if mod not in self._known_modules:
            self._known_modules[mod] = type(mod).__name__
        mod_name = self._known_modules[mod]
        if mod not in self._seen_modules:
            for name, submod in mod.named_children():
                self._known_modules[submod] = f"{mod_name}.{name}"
                self._get_mod_name(submod)
            self._seen_modules.add(mod)
        return mod_name

    def _get_append_fn(self, name, is_bw):
        def fn(*args):
            if is_bw:
                self._maybe_set_engine_callback()
            if name in self.parents:
                print(
                    "The module hierarchy tracking seems to be messed up."
                    "Please file a bug to PyTorch."
                )
            self.parents.add(name)

        return fn

    def _get_pop_fn(self, name, is_bw):
        def fn(*args):
            if name in self.parents:
                self.parents.remove(name)
            elif not is_bw:
                # Due to some input/output not requiring gradients, we cannot enforce
                # proper nesting in backward
                raise RuntimeError(
                    "The Module hierarchy tracking is wrong. Report a bug to PyTorch"
                )

        return fn

    def _fw_pre_hook(self, mod, input):
        name = self._get_mod_name(mod)
        self._get_append_fn(name, False)()

        args, _ = tree_flatten(input)
        tensors = [a for a in args if isinstance(a, torch.Tensor) and a.requires_grad]
        if tensors:
            register_multi_grad_hook(tensors, self._get_pop_fn(name, True))

    def _fw_post_hook(self, mod, input, output):
        name = self._get_mod_name(mod)
        self._get_pop_fn(name, False)()

        args, _ = tree_flatten(output)
        tensors = [a for a in args if isinstance(a, torch.Tensor) and a.requires_grad]
        if tensors:
            register_multi_grad_hook(tensors, self._get_append_fn(name, True))

    def __enter__(self):
        self._fw_pre_handle = register_module_forward_pre_hook(self._fw_pre_hook)
        self._fw_post_handle = register_module_forward_hook(self._fw_post_hook)
        return self

    def __exit__(self, *args):
        self._fw_pre_handle.remove()
        self._fw_post_handle.remove()



from logging import getLogger
from typing import Any, Literal, Protocol, TypedDict, cast, runtime_checkable

from typing_extensions import NotRequired, override

from ._throughput_monitor_callback import ThroughputMonitor as _ThroughputMonitor
from .base import CallbackConfigBase

log = getLogger(__name__)


class ThroughputMonitorBatchStats(TypedDict):
    batch_size: int
    length: NotRequired[int | None]


@runtime_checkable
class SupportsThroughputMonitorModuleProtocol(Protocol):
    def throughput_monitor_batch_stats(
        self, batch: Any
    ) -> ThroughputMonitorBatchStats: ...


class ThroughputMonitor(_ThroughputMonitor):
    def __init__(self, window_size: int = 100) -> None:
        super().__init__(cast(Any, None), cast(Any, None), window_size=window_size)

    @override
    def setup(self, trainer, pl_module, stage):
        if not isinstance(pl_module, SupportsThroughputMonitorModuleProtocol):
            raise RuntimeError(
                "The model does not implement `throughput_monitor_batch_stats`. "
                "Please either implement this method, or do not use the `ThroughputMonitor` callback."
            )

        def batch_size_fn(batch):
            return pl_module.throughput_monitor_batch_stats(batch)["batch_size"]

        def length_fn(batch):
            return pl_module.throughput_monitor_batch_stats(batch).get("length")

        self.batch_size_fn = batch_size_fn
        self.length_fn = length_fn

        return super().setup(trainer, pl_module, stage)


class ThroughputMonitorConfig(CallbackConfigBase):
    name: Literal["throughput_monitor"] = "throughput_monitor"

    window_size: int = 100
    """Number of batches to use for a rolling average."""

    @override
    def create_callbacks(self, root_config):
        yield ThroughputMonitor(window_size=self.window_size)



import re
from datetime import timedelta
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING, Literal

from lightning.pytorch.callbacks.model_checkpoint import (
    ModelCheckpoint as _ModelCheckpoint,
)
from typing_extensions import override

from ..metrics import MetricConfig
from .base import CallbackConfigBase

if TYPE_CHECKING:
    from ..model.config import BaseConfig

log = getLogger(__name__)


def _convert_string(input_string: str):
    # Find all variables enclosed in curly braces
    variables = re.findall(r"\{(.*?)\}", input_string)

    # Replace each variable with its corresponding key-value pair
    output_string = input_string
    for variable in variables:
        # If the name is something like {variable:format}, we shouldn't process the format.
        key_name = variable
        if ":" in variable:
            key_name, _ = variable.split(":", 1)
            continue

        # Replace '/' with '_' in the key name
        key_name = key_name.replace("/", "_")
        output_string = output_string.replace(
            f"{{{variable}}}", f"{key_name}={{{variable}}}"
        )

    return output_string


class ModelCheckpointCallbackConfig(CallbackConfigBase):
    """Arguments for the ModelCheckpoint callback."""

    kind: Literal["model_checkpoint"] = "model_checkpoint"

    dirpath: str | Path | None = None
    """
    Directory path to save the model file. If `None`, we save to the checkpoint directory set in `config.directory`.
    """

    filename: str | None = None
    """
    Checkpoint filename.
        If None, a default template is used (see :attr:`ModelCheckpoint.CHECKPOINT_JOIN_CHAR`).
    """

    metric: MetricConfig | None = None
    """
    Metric to monitor for saving checkpoints.
        If None, the primary metric of the runner will be used, if available.
    """

    verbose: bool = False
    """Verbosity mode. If True, print additional information about checkpoints."""

    save_last: Literal[True, False, "link"] | None = "link"
    """
    Whether to save the last checkpoint.
        If True, saves a copy of the last checkpoint separately.
        If "link", creates a symbolic link to the last checkpoint.
    """

    save_top_k: int = 1
    """
    Number of best models to save.
        If -1, all models are saved.
        If 0, no models are saved.
    """

    save_weights_only: bool = False
    """Whether to save only the model's weights or the entire model object."""

    auto_insert_metric_name: bool = True
    """Whether to automatically insert the metric name in the checkpoint filename."""

    every_n_train_steps: int | None = None
    """
    Number of training steps between checkpoints.
        If None or 0, no checkpoints are saved during training.
    """

    train_time_interval: timedelta | None = None
    """
    Time interval between checkpoints during training.
        If None, no checkpoints are saved during training based on time.
    """

    every_n_epochs: int | None = None
    """
    Number of epochs between checkpoints.
        If None or 0, no checkpoints are saved at the end of epochs.
    """

    save_on_train_epoch_end: bool | None = None
    """
    Whether to run checkpointing at the end of the training epoch.
        If False, checkpointing runs at the end of the validation.
    """

    enable_version_counter: bool = True
    """Whether to append a version to the existing file name."""

    auto_append_metric: bool = True
    """If enabled, this will automatically add "-{monitor}" to the filename."""

    def metric_or_default(self, root_config: "BaseConfig"):
        if self.metric is not None:
            return self.metric
        if root_config.primary_metric is not None:
            return root_config.primary_metric
        raise ValueError("Primary metric must be provided if metric is not specified.")

    def resolve_filename(self, root_config: "BaseConfig"):
        metric = self.metric_or_default(root_config)

        filename = self.filename
        if not filename:
            filename = "{epoch}-{step}"
        if self.auto_append_metric:
            filename = f"{filename}-{{{metric.validation_monitor}}}"

        if self.auto_insert_metric_name and filename:
            new_filename = _convert_string(filename)
            log.critical(
                f"Updated ModelCheckpoint filename: {filename} -> {new_filename}"
            )
            filename = new_filename

        return filename

    @override
    def create_callbacks(self, root_config):
        dirpath = self.dirpath or root_config.directory.resolve_subdirectory(
            root_config.id, "checkpoint"
        )

        metric = self.metric_or_default(root_config)
        filename = self.resolve_filename(root_config)

        yield ModelCheckpoint(
            self,
            dirpath=Path(dirpath),
            filename=filename,
            metric=metric,
        )


class ModelCheckpoint(_ModelCheckpoint):
    @override
    def __init__(
        self,
        config: ModelCheckpointCallbackConfig,
        dirpath: Path,
        filename: str,
        metric: MetricConfig,
    ):
        self.config = config
        del config

        super().__init__(
            dirpath=dirpath,
            filename=filename,
            monitor=metric.validation_monitor,
            mode=metric.mode,
            verbose=self.config.verbose,
            save_last=self.config.save_last,
            save_top_k=self.config.save_top_k,
            save_weights_only=self.config.save_weights_only,
            auto_insert_metric_name=False,
            every_n_train_steps=self.config.every_n_train_steps,
            train_time_interval=self.config.train_time_interval,
            every_n_epochs=self.config.every_n_epochs,
            save_on_train_epoch_end=self.config.save_on_train_epoch_end,
            enable_version_counter=self.config.enable_version_counter,
        )



from lightning.pytorch import LightningDataModule, LightningModule
from lightning.pytorch.profilers import PassThroughProfiler

from ...util.typing_utils import mixin_base_type


class ProfilerMixin(mixin_base_type(LightningModule)):
    @property
    def profiler(self):
        if not isinstance(self, (LightningModule, LightningDataModule)):
            raise TypeError(
                "`profiler` can only be used on LightningModule or LightningDataModule"
            )

        if (trainer := self.trainer) is None:
            raise RuntimeError("trainer is not defined")

        if not hasattr(trainer, "profiler"):
            raise RuntimeError("trainer does not have profiler")

        if (profiler := getattr(trainer, "profiler")) is None:
            profiler = PassThroughProfiler()

        return profiler



from . import _experimental as _experimental
from . import callbacks as callbacks
from . import data as data
from . import lr_scheduler as lr_scheduler
from . import metrics as metrics
from . import model as model
from . import nn as nn
from . import optimizer as optimizer
from .metrics import MetricConfig as MetricConfig
from .model import Base as Base
from .model import BaseConfig as BaseConfig
from .model import ConfigList as ConfigList
from .model import LightningModuleBase as LightningModuleBase
from .runner import Runner as Runner
from .trainer import Trainer as Trainer



import copy
import functools
from collections.abc import Callable, Iterable, Mapping, Sequence
from typing import Generic

from nshrunner import RunInfo
from nshrunner import Runner as _Runner
from nshrunner._submit import screen
from nshrunner.snapshot import SnapshotArgType
from typing_extensions import TypeVar, TypeVarTuple, Unpack, override

from .model.config import BaseConfig

TConfig = TypeVar("TConfig", bound=BaseConfig, infer_variance=True)
TArguments = TypeVarTuple("TArguments", default=Unpack[tuple[()]])
TReturn = TypeVar("TReturn", infer_variance=True)


class Runner(
    _Runner[TReturn, TConfig, Unpack[TArguments]],
    Generic[TReturn, TConfig, Unpack[TArguments]],
):
    @override
    @classmethod
    def default_validate_fn(cls, config: TConfig, *args: Unpack[TArguments]) -> None:
        super().default_validate_fn(config, *args)

    @override
    @classmethod
    def default_info_fn(cls, config: TConfig, *args: Unpack[TArguments]) -> RunInfo:
        run_info = super().default_info_fn(config, *args)
        return {
            **run_info,
            "id": config.id,
            "base_dir": config.directory.project_root,
        }

    def _fast_dev_run_transform(
        self,
        config: TConfig,
        *args: Unpack[TArguments],
        n_batches: int,
    ):
        config = copy.deepcopy(config)
        config.trainer.fast_dev_run = n_batches
        return (config, *args)

    def fast_dev_run(
        self,
        runs: Iterable[tuple[TConfig, Unpack[TArguments]]],
        n_batches: int = 1,
        *,
        env: Mapping[str, str] | None = None,
        transforms: list[
            Callable[[TConfig, Unpack[TArguments]], tuple[TConfig, Unpack[TArguments]]]
        ]
        | None = None,
    ):
        transforms = transforms or []
        transforms.append(
            functools.partial(self._fast_dev_run_transform, n_batches=n_batches)
        )
        return self.local(
            runs,
            env=env,
            transforms=transforms,
        )

    def fast_dev_run_generator(
        self,
        runs: Iterable[tuple[TConfig, Unpack[TArguments]]],
        n_batches: int = 1,
        *,
        env: Mapping[str, str] | None = None,
        transforms: list[
            Callable[[TConfig, Unpack[TArguments]], tuple[TConfig, Unpack[TArguments]]]
        ]
        | None = None,
    ):
        transforms = transforms or []
        transforms.append(
            functools.partial(self._fast_dev_run_transform, n_batches=n_batches)
        )
        return self.local_generator(
            runs,
            env=env,
            transforms=transforms,
        )

    def fast_dev_run_session(
        self,
        runs: Iterable[tuple[TConfig, Unpack[TArguments]]],
        options: screen.ScreenJobKwargs = {},
        n_batches: int = 1,
        *,
        snapshot: SnapshotArgType,
        setup_commands: Sequence[str] | None = None,
        env: Mapping[str, str] | None = None,
        transforms: list[
            Callable[[TConfig, Unpack[TArguments]], tuple[TConfig, Unpack[TArguments]]]
        ]
        | None = None,
        activate_venv: bool = True,
        print_command: bool = True,
    ):
        transforms = transforms or []
        transforms.append(
            functools.partial(self._fast_dev_run_transform, n_batches=n_batches)
        )
        return self.session(
            runs,
            options,
            snapshot=snapshot,
            setup_commands=setup_commands,
            env=env,
            transforms=transforms,
            activate_venv=activate_venv,
            print_command=print_command,
        )



from typing import Annotated, TypeAlias

import nshconfig as C

from ._base import LRSchedulerConfigBase as LRSchedulerConfigBase
from ._base import LRSchedulerMetadata as LRSchedulerMetadata
from .linear_warmup_cosine import (
    LinearWarmupCosineAnnealingLR as LinearWarmupCosineAnnealingLR,
)
from .linear_warmup_cosine import (
    LinearWarmupCosineDecayLRSchedulerConfig as LinearWarmupCosineDecayLRSchedulerConfig,
)
from .reduce_lr_on_plateau import ReduceLROnPlateau as ReduceLROnPlateau
from .reduce_lr_on_plateau import ReduceLROnPlateauConfig as ReduceLROnPlateauConfig

LRSchedulerConfig: TypeAlias = Annotated[
    LinearWarmupCosineDecayLRSchedulerConfig | ReduceLROnPlateauConfig,
    C.Field(discriminator="name"),
]



import math
from logging import getLogger

from lightning.fabric.utilities.rank_zero import _get_rank
from lightning.pytorch import Trainer
from lightning.pytorch.callbacks import EarlyStopping as _EarlyStopping
from lightning.pytorch.utilities.rank_zero import rank_prefixed_message
from typing_extensions import override

log = getLogger(__name__)


class EarlyStopping(_EarlyStopping):
    def __init__(
        self,
        monitor: str,
        min_delta: float = 0,
        min_lr: float | None = None,
        patience: int = 3,
        verbose: bool = True,
        mode: str = "min",
        strict: bool = True,
        check_finite: bool = True,
        stopping_threshold: float | None = None,
        divergence_threshold: float | None = None,
        check_on_train_epoch_end: bool | None = None,
        log_rank_zero_only: bool = False,
    ):
        super().__init__(
            monitor,
            min_delta,
            patience,
            verbose,
            mode,
            strict,
            check_finite,
            stopping_threshold,
            divergence_threshold,
            check_on_train_epoch_end,
            log_rank_zero_only,
        )

        self.min_lr = min_lr

    @override
    @staticmethod
    def _log_info(
        trainer: Trainer | None, message: str, log_rank_zero_only: bool
    ) -> None:
        rank = _get_rank()
        if trainer is not None and trainer.world_size <= 1:
            rank = None
        message = rank_prefixed_message(message, rank)
        if rank is None or not log_rank_zero_only or rank == 0:
            log.critical(message)

    @override
    def _run_early_stopping_check(self, trainer: Trainer):
        """Checks whether the early stopping condition is met and if so tells the trainer to stop the training."""
        logs = trainer.callback_metrics

        # Disable early_stopping with fast_dev_run
        if getattr(trainer, "fast_dev_run", False):
            return

        should_stop, reason = False, None

        if not should_stop:
            should_stop, reason = self._evaluate_stopping_criteria_min_lr(trainer)

        # If metric present
        if not should_stop and self._validate_condition_metric(logs):
            current = logs[self.monitor].squeeze()
            should_stop, reason = self._evaluate_stopping_criteria(current)

        # stop every ddp process if any world process decides to stop
        should_stop = trainer.strategy.reduce_boolean_decision(should_stop, all=False)
        trainer.should_stop = trainer.should_stop or should_stop
        if should_stop:
            self.stopped_epoch = trainer.current_epoch
        if reason and self.verbose:
            self._log_info(trainer, reason, self.log_rank_zero_only)

    def _evaluate_stopping_criteria_min_lr(
        self, trainer: Trainer
    ) -> tuple[bool, str | None]:
        if self.min_lr is None:
            return False, None

        # Get the maximum LR across all param groups in all optimizers
        model_max_lr = max(
            [
                param_group["lr"]
                for optimizer in trainer.optimizers
                for param_group in optimizer.param_groups
            ]
        )
        if not isinstance(model_max_lr, float) or not math.isfinite(model_max_lr):
            return False, None

        # If the maximum LR is less than the minimum LR, stop training
        if model_max_lr >= self.min_lr:
            return False, None

        return True, (
            "Stopping threshold reached: "
            f"The maximum LR of the model across all param groups is {model_max_lr:.2e} "
            f"which is less than the minimum LR {self.min_lr:.2e}"
        )

    def on_early_stopping(self, trainer: Trainer):
        pass



import builtins
from typing import Literal

import nshconfig as C


class MetricConfig(C.Config):
    name: str
    """The name of the primary metric."""

    mode: Literal["min", "max"]
    """
    The mode of the primary metric:
    - "min" for metrics that should be minimized (e.g., loss)
    - "max" for metrics that should be maximized (e.g., accuracy)
    """

    @property
    def validation_monitor(self) -> str:
        return f"val/{self.name}"

    def __post_init__(self):
        for split in ("train", "val", "test", "predict"):
            if self.name.startswith(f"{split}/"):
                raise ValueError(
                    f"Primary metric name should not start with '{split}/'. "
                    f"Just use '{self.name[len(split) + 1:]}' instead. "
                    "The split name is automatically added depending on the context."
                )

    @classmethod
    def loss(cls, mode: Literal["min", "max"] = "min"):
        return cls(name="loss", mode=mode)

    @property
    def best(self):
        return builtins.min if self.mode == "min" else builtins.max



from collections.abc import Callable

import torch
from lightning.fabric.utilities.imports import _TORCH_GREATER_EQUAL_2_1

MEASURE_FLOPS_AVAILABLE = _TORCH_GREATER_EQUAL_2_1


def measure_flops(
    forward_fn: Callable[[], torch.Tensor],
    loss_fn: Callable[[torch.Tensor], torch.Tensor] | None = None,
    display: bool = True,
) -> int:
    """Utility to compute the total number of FLOPs used by a module during training or during inference.

    It's recommended to create a meta-device model for this:

    Example::

        with torch.device("meta"):
            model = MyModel()
            x = torch.randn(2, 32)

        model_fwd = lambda: model(x)
        fwd_flops = measure_flops(model, model_fwd)

        model_loss = lambda y: y.sum()
        fwd_and_bwd_flops = measure_flops(model, model_fwd, model_loss)

    Args:
        model: The model whose FLOPs should be measured.
        forward_fn: A function that runs ``forward`` on the model and returns the result.
        loss_fn: A function that computes the loss given the ``forward_fn`` output. If provided, the loss and `backward`
            FLOPs will be included in the result.

    """
    if not MEASURE_FLOPS_AVAILABLE:
        raise ImportError("`measure_flops` requires PyTorch >= 2.1.")

    from .flop_counter import FlopCounterMode

    flop_counter = FlopCounterMode(display=display)
    with flop_counter:
        if loss_fn is None:
            forward_fn()
        else:
            loss_fn(forward_fn()).backward()
    return flop_counter.get_total_flops()



from logging import getLogger
from typing import Literal

import torch
from lightning.pytorch import Callback, LightningModule, Trainer
from typing_extensions import override

from .base import CallbackConfigBase

log = getLogger(__name__)


def finite_checks(
    module: LightningModule,
    nonfinite_grads: bool = True,
    none_grads: bool = False,
):
    for name, param in module.named_parameters():
        if not param.requires_grad:
            continue

        if param.grad is None:
            if none_grads:
                log.critical(f"Parameter {name} ({param.shape}) has None gradients")
            continue

        if not nonfinite_grads or torch.isfinite(param.grad.float()).all():
            continue

        has_nan = torch.isnan(param.grad.float()).any()
        has_inf = torch.isinf(param.grad.float()).any()
        kinds = [
            "NaN" if has_nan else None,
            "Inf" if has_inf else None,
        ]
        kinds = " and ".join(prop for prop in kinds if prop is not None)
        log.critical(f"{name} ({param.shape}) has {kinds} gradients")


class FiniteChecksCallback(Callback):
    def __init__(
        self,
        *,
        nonfinite_grads: bool = True,
        none_grads: bool = True,
    ):
        super().__init__()

        self._nonfinite_grads = nonfinite_grads
        self._none_grads = none_grads

    @override
    def on_after_backward(self, trainer: Trainer, pl_module: LightningModule):
        finite_checks(
            pl_module,
            nonfinite_grads=self._nonfinite_grads,
            none_grads=self._none_grads,
        )


class FiniteChecksConfig(CallbackConfigBase):
    name: Literal["finite_checks"] = "finite_checks"

    nonfinite_grads: bool = True
    """Whether to check for non-finite (i.e. NaN or Inf) gradients"""

    none_grads: bool = True
    """Whether to check for None gradients"""

    @override
    def create_callbacks(self, root_config):
        yield FiniteChecksCallback(
            nonfinite_grads=self.nonfinite_grads,
            none_grads=self.none_grads,
        )



from logging import getLogger
from typing import Literal, Protocol, cast, runtime_checkable

import torch.nn as nn
from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks.callback import Callback
from lightning.pytorch.loggers import WandbLogger
from typing_extensions import override

from .base import CallbackConfigBase

log = getLogger(__name__)


@runtime_checkable
class _HasWandbLogModuleProtocol(Protocol):
    def wandb_log_module(self) -> nn.Module | None: ...


class WandbWatchCallback(Callback):
    def __init__(self, config: "WandbWatchConfig"):
        super().__init__()

        self.config = config

    @override
    def on_fit_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self._on_start(trainer, pl_module)

    @override
    def on_validation_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self._on_start(trainer, pl_module)

    @override
    def on_test_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self._on_start(trainer, pl_module)

    @override
    def on_predict_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self._on_start(trainer, pl_module)

    def _on_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        # If not enabled, return
        if not self.config:
            return

        # If we're in fast_dev_run, don't watch the model
        if getattr(trainer, "fast_dev_run", False):
            return

        if (
            logger := next(
                (
                    logger
                    for logger in trainer.loggers
                    if isinstance(logger, WandbLogger)
                ),
                None,
            )
        ) is None:
            log.warning("Could not find wandb logger or module to log")
            return

        if getattr(pl_module, "_model_watched", False):
            return

        # Get which module to log
        if (
            not isinstance(pl_module, _HasWandbLogModuleProtocol)
            or (module := pl_module.wandb_log_module()) is None
        ):
            module = cast(nn.Module, pl_module)

        logger.watch(
            module,
            log=cast(str, self.config.log),
            log_freq=self.config.log_freq,
            log_graph=self.config.log_graph,
        )
        setattr(pl_module, "_model_watched", True)


class WandbWatchConfig(CallbackConfigBase):
    name: Literal["finite_checks"] = "finite_checks"

    enabled: bool = True
    """Enable watching the model for wandb."""

    log: str | None = None
    """Log type for wandb."""

    log_graph: bool = True
    """Whether to log the graph for wandb."""

    log_freq: int = 100
    """Log frequency for wandb."""

    def __bool__(self):
        return self.enabled

    @override
    def create_callbacks(self, root_config):
        yield WandbWatchCallback(self)



REQUIRED_PACKAGES = [
    "beartype",
    "cloudpickle",
    "jaxtyping",
    "lightning",
    "lightning_fabric",
    "lightning_utilities",
    "lovely_numpy",
    "lovely_tensors",
    "numpy",
    "psutil",
    "pydantic",
    "pydantic_core",
    "pysnooper",
    "rich",
    "tabulate",
    "torch",
    "torchmetrics",
    "tqdm",
    "typing_extensions",
    "wrapt",
    "yaml",
]


def main():
    import importlib.util
    import sys

    missing_packages: list[str] = []
    for package_name in REQUIRED_PACKAGES:
        spec = importlib.util.find_spec(package_name)
        if spec is None:
            missing_packages.append(package_name)

    if missing_packages:
        sys.exit(f"Error: Missing required packages: {', '.join(missing_packages)}")


if __name__ == "__main__":
    main()



import copy
import datetime
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import nshconfig as C
import numpy as np
import torch

from ..model._environment import EnvironmentConfig

if TYPE_CHECKING:
    from ..model import BaseConfig, LightningModuleBase
    from .trainer import Trainer

log = logging.getLogger(__name__)


METADATA_PATH_SUFFIX = ".metadata.json"
HPARAMS_PATH_SUFFIX = ".hparams.json"


class CheckpointMetadata(C.Config):
    checkpoint_path: Path
    checkpoint_filename: str

    run_id: str
    name: str
    project: str | None
    checkpoint_timestamp: datetime.datetime
    start_timestamp: datetime.datetime | None

    epoch: int
    global_step: int
    training_time: datetime.timedelta
    metrics: dict[str, Any]
    environment: EnvironmentConfig

    @classmethod
    def from_file(cls, path: Path):
        return cls.model_validate_json(path.read_text())


def _generate_checkpoint_metadata(
    config: "BaseConfig", trainer: "Trainer", checkpoint_path: Path
):
    checkpoint_timestamp = datetime.datetime.now()
    start_timestamp = trainer.start_time()
    training_time = trainer.time_elapsed()

    metrics: dict[str, Any] = {}
    for name, metric in copy.deepcopy(trainer.callback_metrics).items():
        match metric:
            case torch.Tensor() | np.ndarray():
                metrics[name] = metric.detach().cpu().item()
            case _:
                metrics[name] = metric

    return CheckpointMetadata(
        checkpoint_path=checkpoint_path,
        checkpoint_filename=checkpoint_path.name,
        run_id=config.id,
        name=config.run_name,
        project=config.project,
        checkpoint_timestamp=checkpoint_timestamp,
        start_timestamp=start_timestamp.datetime
        if start_timestamp is not None
        else None,
        epoch=trainer.current_epoch,
        global_step=trainer.global_step,
        training_time=training_time,
        metrics=metrics,
        environment=config.environment,
    )


def _write_checkpoint_metadata(
    trainer: "Trainer",
    model: "LightningModuleBase",
    checkpoint_path: Path,
):
    config = cast("BaseConfig", model.config)
    metadata = _generate_checkpoint_metadata(config, trainer, checkpoint_path)

    # Write the metadata to the checkpoint directory
    try:
        metadata_path = checkpoint_path.with_suffix(METADATA_PATH_SUFFIX)
        metadata_path.write_text(metadata.model_dump_json(indent=4))
    except Exception as e:
        log.warning(f"Failed to write metadata to {checkpoint_path}: {e}")
    else:
        log.info(f"Checkpoint metadata written to {checkpoint_path}")

    # Write the hparams to the checkpoint directory
    try:
        hparams_path = checkpoint_path.with_suffix(HPARAMS_PATH_SUFFIX)
        hparams_path.write_text(config.model_dump_json(indent=4))
    except Exception as e:
        log.warning(f"Failed to write hparams to {checkpoint_path}: {e}")
    else:
        log.info(f"Checkpoint metadata written to {checkpoint_path}")



from collections import abc
from collections.abc import Callable, Iterable
from logging import getLogger
from typing import Any, TypeAlias, cast, final

from lightning.pytorch import Callback, LightningModule
from lightning.pytorch.callbacks import LambdaCallback
from typing_extensions import override

from ...util.typing_utils import mixin_base_type

log = getLogger(__name__)

CallbackFn: TypeAlias = Callable[[], Callback | Iterable[Callback] | None]


class CallbackRegistrarModuleMixin:
    @override
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._ll_callbacks: list[CallbackFn] = []

    def register_callback(
        self,
        callback: Callback | Iterable[Callback] | CallbackFn | None = None,
        *,
        setup: Callable | None = None,
        teardown: Callable | None = None,
        on_fit_start: Callable | None = None,
        on_fit_end: Callable | None = None,
        on_sanity_check_start: Callable | None = None,
        on_sanity_check_end: Callable | None = None,
        on_train_batch_start: Callable | None = None,
        on_train_batch_end: Callable | None = None,
        on_train_epoch_start: Callable | None = None,
        on_train_epoch_end: Callable | None = None,
        on_validation_epoch_start: Callable | None = None,
        on_validation_epoch_end: Callable | None = None,
        on_test_epoch_start: Callable | None = None,
        on_test_epoch_end: Callable | None = None,
        on_validation_batch_start: Callable | None = None,
        on_validation_batch_end: Callable | None = None,
        on_test_batch_start: Callable | None = None,
        on_test_batch_end: Callable | None = None,
        on_train_start: Callable | None = None,
        on_train_end: Callable | None = None,
        on_validation_start: Callable | None = None,
        on_validation_end: Callable | None = None,
        on_test_start: Callable | None = None,
        on_test_end: Callable | None = None,
        on_exception: Callable | None = None,
        on_save_checkpoint: Callable | None = None,
        on_load_checkpoint: Callable | None = None,
        on_before_backward: Callable | None = None,
        on_after_backward: Callable | None = None,
        on_before_optimizer_step: Callable | None = None,
        on_before_zero_grad: Callable | None = None,
        on_predict_start: Callable | None = None,
        on_predict_end: Callable | None = None,
        on_predict_batch_start: Callable | None = None,
        on_predict_batch_end: Callable | None = None,
        on_predict_epoch_start: Callable | None = None,
        on_predict_epoch_end: Callable | None = None,
    ):
        if callback is None:
            callback = LambdaCallback(
                setup=setup,
                teardown=teardown,
                on_fit_start=on_fit_start,
                on_fit_end=on_fit_end,
                on_sanity_check_start=on_sanity_check_start,
                on_sanity_check_end=on_sanity_check_end,
                on_train_batch_start=on_train_batch_start,
                on_train_batch_end=on_train_batch_end,
                on_train_epoch_start=on_train_epoch_start,
                on_train_epoch_end=on_train_epoch_end,
                on_validation_epoch_start=on_validation_epoch_start,
                on_validation_epoch_end=on_validation_epoch_end,
                on_test_epoch_start=on_test_epoch_start,
                on_test_epoch_end=on_test_epoch_end,
                on_validation_batch_start=on_validation_batch_start,
                on_validation_batch_end=on_validation_batch_end,
                on_test_batch_start=on_test_batch_start,
                on_test_batch_end=on_test_batch_end,
                on_train_start=on_train_start,
                on_train_end=on_train_end,
                on_validation_start=on_validation_start,
                on_validation_end=on_validation_end,
                on_test_start=on_test_start,
                on_test_end=on_test_end,
                on_exception=on_exception,
                on_save_checkpoint=on_save_checkpoint,
                on_load_checkpoint=on_load_checkpoint,
                on_before_backward=on_before_backward,
                on_after_backward=on_after_backward,
                on_before_optimizer_step=on_before_optimizer_step,
                on_before_zero_grad=on_before_zero_grad,
                on_predict_start=on_predict_start,
                on_predict_end=on_predict_end,
                on_predict_batch_start=on_predict_batch_start,
                on_predict_batch_end=on_predict_batch_end,
                on_predict_epoch_start=on_predict_epoch_start,
                on_predict_epoch_end=on_predict_epoch_end,
            )

        if not callable(callback):
            callback_ = cast(CallbackFn, lambda: callback)
        else:
            callback_ = callback

        self._ll_callbacks.append(callback_)


class CallbackModuleMixin(
    CallbackRegistrarModuleMixin,
    mixin_base_type(LightningModule),
):
    def _gather_all_callbacks(self):
        modules: list[Any] = []
        if isinstance(self, CallbackRegistrarModuleMixin):
            modules.append(self)
        if (
            datamodule := getattr(self.trainer, "datamodule", None)
        ) is not None and isinstance(datamodule, CallbackRegistrarModuleMixin):
            modules.append(datamodule)
        modules.extend(
            module
            for module in self.children()
            if isinstance(module, CallbackRegistrarModuleMixin)
        )
        for module in modules:
            yield from module._ll_callbacks

    @final
    @override
    def configure_callbacks(self):
        callbacks = super().configure_callbacks()
        if not isinstance(callbacks, abc.Sequence):
            callbacks = [callbacks]

        callbacks = list(callbacks)
        for callback_fn in self._gather_all_callbacks():
            callback_result = callback_fn()
            if callback_result is None:
                continue

            if not isinstance(callback_result, abc.Iterable):
                callback_result = [callback_result]

            for callback in callback_result:
                log.info(
                    f"Registering {callback.__class__.__qualname__} callback {callback}"
                )
                callbacks.append(callback)

        return callbacks



from collections.abc import Mapping
from logging import getLogger
from typing import cast

import torch
from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.utilities.types import (
    LRSchedulerConfigType,
    LRSchedulerTypeUnion,
)
from typing_extensions import Protocol, override, runtime_checkable

from ...util.typing_utils import mixin_base_type
from ..config import BaseConfig
from .callback import CallbackModuleMixin

log = getLogger(__name__)


def _on_train_start_callback(trainer: Trainer, pl_module: LightningModule):
    # If we're in PL's "sanity check" mode, we don't need to run this check
    if trainer.sanity_checking:
        return

    config = cast(BaseConfig, pl_module.hparams)
    if config.trainer.sanity_checking.reduce_lr_on_plateau == "disable":
        return

    # if no lr schedulers, return
    if not trainer.lr_scheduler_configs:
        return

    errors: list[str] = []
    disable_message = (
        "Otherwise, set `config.trainer.sanity_checking.reduce_lr_on_plateau='disable'` "
        "to disable this sanity check."
    )

    for lr_scheduler_config in trainer.lr_scheduler_configs:
        if not lr_scheduler_config.reduce_on_plateau:
            continue

        match lr_scheduler_config.interval:
            case "epoch":
                # we need to make sure that the trainer runs val every `frequency` epochs

                # If `trainer.check_val_every_n_epoch` is None, then Lightning
                # will run val every `int(trainer.val_check_interval)` steps.
                # So, first we need to make sure that `trainer.val_check_interval` is not None first.
                if trainer.check_val_every_n_epoch is None:
                    errors.append(
                        "Trainer is not running validation at epoch intervals "
                        "(i.e., `trainer.check_val_every_n_epoch` is None) but "
                        f"a ReduceLRPlateau scheduler with interval={lr_scheduler_config.interval} is used."
                        f"Please set `config.trainer.check_val_every_n_epoch={lr_scheduler_config.frequency}`. "
                        + disable_message
                    )

                # Second, we make sure that the trainer runs val at least every `frequency` epochs
                if (
                    trainer.check_val_every_n_epoch is not None
                    and lr_scheduler_config.frequency % trainer.check_val_every_n_epoch
                    != 0
                ):
                    errors.append(
                        f"Trainer is not running validation every {lr_scheduler_config.frequency} epochs but "
                        f"a ReduceLRPlateau scheduler with interval={lr_scheduler_config.interval} and frequency={lr_scheduler_config.frequency} is used."
                        f"Please set `config.trainer.check_val_every_n_epoch` to a multiple of {lr_scheduler_config.frequency}. "
                        + disable_message
                    )

            case "step":
                # In this case, we need to make sure that the trainer runs val at step intervals
                # that are multiples of `frequency`.

                # First, we make sure that validation is run at step intervals
                if trainer.check_val_every_n_epoch is not None:
                    errors.append(
                        "Trainer is running validation at epoch intervals "
                        "(i.e., `trainer.check_val_every_n_epoch` is not None) but "
                        f"a ReduceLRPlateau scheduler with interval={lr_scheduler_config.interval} is used."
                        "Please set `config.trainer.check_val_every_n_epoch=None` "
                        f"and `config.trainer.val_check_interval={lr_scheduler_config.frequency}`. "
                        + disable_message
                    )

                # Second, we make sure `trainer.val_check_interval` is an integer
                if not isinstance(trainer.val_check_interval, int):
                    errors.append(
                        f"Trainer is not running validation at step intervals "
                        f"(i.e., `trainer.val_check_interval` is not an integer) but "
                        f"a ReduceLRPlateau scheduler with interval={lr_scheduler_config.interval} is used."
                        "Please set `config.trainer.val_check_interval=None` "
                        f"and `config.trainer.val_check_interval={lr_scheduler_config.frequency}`. "
                        + disable_message
                    )

                # Third, we make sure that the trainer runs val at least every `frequency` steps
                if (
                    isinstance(trainer.val_check_interval, int)
                    and trainer.val_check_interval % lr_scheduler_config.frequency != 0
                ):
                    errors.append(
                        f"Trainer is not running validation every {lr_scheduler_config.frequency} steps but "
                        f"a ReduceLRPlateau scheduler with interval={lr_scheduler_config.interval} and frequency={lr_scheduler_config.frequency} is used."
                        "Please set `config.trainer.val_check_interval` "
                        f"to a multiple of {lr_scheduler_config.frequency}. "
                        + disable_message
                    )

            case _:
                pass

    if not errors:
        return

    message = (
        "ReduceLRPlateau sanity checks failed with the following errors:\n"
        + "\n".join(errors)
    )
    match config.trainer.sanity_checking.reduce_lr_on_plateau:
        case "warn":
            log.warning(message)
        case "error":
            raise ValueError(message)
        case _:
            pass


@runtime_checkable
class CustomRLPImplementation(Protocol):
    __reduce_lr_on_plateau__: bool


class RLPSanityCheckModuleMixin(mixin_base_type(CallbackModuleMixin)):
    @override
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        global _on_train_start_callback
        self.register_callback(on_train_start=_on_train_start_callback)

    def reduce_lr_on_plateau_config(
        self,
        lr_scheduler: LRSchedulerTypeUnion | LRSchedulerConfigType,
    ) -> LRSchedulerConfigType:
        if (trainer := self._trainer) is None:
            raise RuntimeError(
                "Could not determine the frequency of ReduceLRPlateau scheduler "
                "because `self.trainer` is None."
            )

        # First, resolve the LR scheduler from the provided config.
        lr_scheduler_config: LRSchedulerConfigType
        match lr_scheduler:
            case Mapping():
                lr_scheduler_config = cast(LRSchedulerConfigType, lr_scheduler)
            case _:
                lr_scheduler_config = {"scheduler": lr_scheduler}

        # Make sure the scheduler is a ReduceLRPlateau scheduler. Otherwise, warn the user.
        if (
            not isinstance(
                lr_scheduler_config["scheduler"],
                torch.optim.lr_scheduler.ReduceLROnPlateau,
            )
        ) and (
            not isinstance(lr_scheduler_config["scheduler"], CustomRLPImplementation)
            or not lr_scheduler_config["scheduler"].__reduce_lr_on_plateau__
        ):
            log.warning(
                "`reduce_lr_on_plateau_config` should only be used with a ReduceLRPlateau scheduler. "
                f"The provided scheduler, {lr_scheduler_config['scheduler']}, does not subclass "
                "`torch.optim.lr_scheduler.ReduceLROnPlateau`. "
                "Please ensure that the scheduler is a ReduceLRPlateau scheduler. "
                "If you are using a custom ReduceLRPlateau scheduler implementation, "
                "please either (1) make sure that it subclasses `torch.optim.lr_scheduler.ReduceLROnPlateau`, "
                "or (2) set the scheduler's `__reduce_lr_on_plateau__` attribute to `True`."
            )

        # If trainer.check_val_every_n_epoch is an integer, then we run val at epoch intervals.
        if trainer.check_val_every_n_epoch is not None:
            return {
                "reduce_on_plateau": True,
                "interval": "epoch",
                "frequency": trainer.check_val_every_n_epoch,
                **lr_scheduler_config,
            }

        # Otherwise, we run val at step intervals.
        if not isinstance(trainer.val_check_batch, int):
            raise ValueError(
                "Could not determine the frequency of ReduceLRPlateau scheduler "
                f"because {trainer.val_check_batch=} is not an integer."
            )

        return {
            "reduce_on_plateau": True,
            "interval": "step",
            "frequency": trainer.val_check_batch,
            **lr_scheduler_config,
        }



from typing_extensions import TypeAlias

from ._environment import (
    EnvironmentClassInformationConfig as EnvironmentClassInformationConfig,
)
from ._environment import EnvironmentConfig as EnvironmentConfig
from ._environment import (
    EnvironmentLinuxEnvironmentConfig as EnvironmentLinuxEnvironmentConfig,
)
from ._environment import (
    EnvironmentSLURMInformationConfig as EnvironmentSLURMInformationConfig,
)
from ._environment import EnvironmentSnapshotConfig as EnvironmentSnapshotConfig
from .base import Base as Base
from .base import LightningModuleBase as LightningModuleBase
from .config import BaseConfig as BaseConfig
from .config import BaseLoggerConfig as BaseLoggerConfig
from .config import BaseProfilerConfig as BaseProfilerConfig
from .config import CheckpointLoadingConfig as CheckpointLoadingConfig
from .config import CheckpointSavingConfig as CheckpointSavingConfig
from .config import DirectoryConfig as DirectoryConfig
from .config import EarlyStoppingConfig as EarlyStoppingConfig
from .config import GradientClippingConfig as GradientClippingConfig
from .config import (
    LatestEpochCheckpointCallbackConfig as LatestEpochCheckpointCallbackConfig,
)
from .config import LoggingConfig as LoggingConfig
from .config import MetricConfig as MetricConfig
from .config import ModelCheckpointCallbackConfig as ModelCheckpointCallbackConfig
from .config import (
    OnExceptionCheckpointCallbackConfig as OnExceptionCheckpointCallbackConfig,
)
from .config import OptimizationConfig as OptimizationConfig
from .config import PrimaryMetricConfig as PrimaryMetricConfig
from .config import ReproducibilityConfig as ReproducibilityConfig
from .config import SanityCheckingConfig as SanityCheckingConfig
from .config import TrainerConfig as TrainerConfig
from .config import WandbWatchConfig as WandbWatchConfig

ConfigList: TypeAlias = list[tuple[BaseConfig, type[LightningModuleBase]]]



from typing import Any, Literal, cast

import torch.distributed
from lightning.pytorch import LightningModule
from torch.distributed import ReduceOp
from typing_extensions import TypeVar

from ...util.typing_utils import mixin_base_type

T = TypeVar("T", infer_variance=True)

ReduceOpStr = Literal[
    "avg",
    "mean",
    "band",
    "bor",
    "bxor",
    "max",
    "min",
    "premul_sum",
    "product",
    "sum",
]
VALID_REDUCE_OPS = (
    "avg",
    "mean",
    "band",
    "bor",
    "bxor",
    "max",
    "min",
    "premul_sum",
    "product",
    "sum",
)


class DistributedMixin(mixin_base_type(LightningModule)):
    def all_gather_object(
        self,
        object: T,
        group: torch.distributed.ProcessGroup | None = None,
    ) -> list[T]:
        if (
            not torch.distributed.is_available()
            or not torch.distributed.is_initialized()
        ):
            return [object]

        object_list = [cast(T, None) for _ in range(self.trainer.world_size)]
        torch.distributed.all_gather_object(object_list, object, group=group)
        return object_list

    def barrier(self, name: str | None = None):
        self.trainer.strategy.barrier(name=name)

    def reduce(
        self,
        tensor: torch.Tensor,
        reduce_op: ReduceOp.RedOpType | ReduceOpStr,
        group: Any | None = None,
    ) -> torch.Tensor:
        if isinstance(reduce_op, str):
            # validate reduce_op
            if reduce_op not in VALID_REDUCE_OPS:
                raise ValueError(
                    f"reduce_op must be one of {VALID_REDUCE_OPS}, got {reduce_op}"
                )

        return self.trainer.strategy.reduce(tensor, group=group, reduce_op=reduce_op)



import logging
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Literal, TypeAlias, overload

import nshconfig as C
from lightning.pytorch import Trainer as LightningTrainer
from lightning.pytorch.trainer.states import TrainerFn
from typing_extensions import assert_never

from ..metrics._config import MetricConfig
from ._checkpoint_metadata import METADATA_PATH_SUFFIX, CheckpointMetadata

if TYPE_CHECKING:
    from ..model.config import BaseConfig

log = logging.getLogger(__name__)


class BestCheckpointStrategyConfig(C.Config):
    name: Literal["best"] = "best"

    metric: MetricConfig | None = None
    """The metric to use for selecting the best checkpoint. If `None`, the primary metric will be used."""

    additional_candidates: Iterable[Path] = []
    """Additional checkpoint candidates to consider when selecting the last checkpoint."""


class UserProvidedPathCheckpointStrategyConfig(C.Config):
    name: Literal["user_provided_path"] = "user_provided_path"

    path: Path
    """The path to the checkpoint to load."""

    on_error: Literal["warn", "raise"] = "warn"
    """The behavior when the checkpoint does not belong to the current run.

    - `warn`: Log a warning and skip the checkpoint.
    - `raise`: Raise an error.
    """


class LastCheckpointStrategyConfig(C.Config):
    name: Literal["last"] = "last"

    criterion: Literal["global_step", "runtime"] = "global_step"
    """The criterion to use for selecting the last checkpoint.

    - `global_step`: The checkpoint with the highest global step will be selected.
    - `runtime`: The checkpoint with the highest runtime will be selected.
    """

    additional_candidates: Iterable[Path] = []
    """Additional checkpoint candidates to consider when selecting the last checkpoint."""


CheckpointLoadingStrategyConfig: TypeAlias = Annotated[
    BestCheckpointStrategyConfig
    | LastCheckpointStrategyConfig
    | UserProvidedPathCheckpointStrategyConfig,
    C.Field(discriminator="name"),
]


class CheckpointLoadingConfig(C.Config):
    strategies: Sequence[CheckpointLoadingStrategyConfig]
    """The strategies to use for loading checkpoints.

    The order of the strategies determines the priority of the strategies.
    The first strategy that resolves a checkpoint will be used.
    """

    include_hpc: bool
    """Whether to include checkpoints from HPC pre-emption."""

    @classmethod
    def _auto_train(cls, ckpt: Literal["best", "last"] | str | Path | None):
        if ckpt is None:
            ckpt = "last"
        match ckpt:
            case "best":
                return cls(
                    strategies=[BestCheckpointStrategyConfig()],
                    include_hpc=True,
                )
            case "last":
                return cls(
                    strategies=[LastCheckpointStrategyConfig()],
                    include_hpc=True,
                )
            case Path() | str():
                ckpt = Path(ckpt)
                return cls(
                    strategies=[
                        LastCheckpointStrategyConfig(additional_candidates=[ckpt]),
                        UserProvidedPathCheckpointStrategyConfig(path=ckpt),
                    ],
                    include_hpc=True,
                )
            case _:
                assert_never(ckpt)

    @classmethod
    def _auto_eval(cls, ckpt: Literal["best", "last"] | str | Path | None):
        if ckpt is None:
            raise ValueError("Checkpoint path must be provided for evaluation.")

        match ckpt:
            case "best":
                return cls(
                    strategies=[BestCheckpointStrategyConfig()],
                    include_hpc=False,
                )
            case "last":
                return cls(
                    strategies=[LastCheckpointStrategyConfig()],
                    include_hpc=False,
                )
            case Path() | str():
                ckpt = Path(ckpt)
                return cls(
                    strategies=[UserProvidedPathCheckpointStrategyConfig(path=ckpt)],
                    include_hpc=False,
                )
            case _:
                assert_never(ckpt)

    @classmethod
    def auto(
        cls,
        ckpt: Literal["best", "last"] | str | Path | None,
        trainer_mode: TrainerFn,
    ):
        match trainer_mode:
            case TrainerFn.FITTING:
                return cls._auto_train(ckpt)
            case TrainerFn.VALIDATING | TrainerFn.TESTING | TrainerFn.PREDICTING:
                return cls._auto_eval(ckpt)
            case _:
                assert_never(trainer_mode)


@dataclass
class _CkptCandidate:
    meta: CheckpointMetadata
    meta_path: Path

    @property
    def ckpt_path(self):
        return self.meta_path.with_name(self.meta.checkpoint_filename)


@overload
def _load_ckpt_meta(
    path: Path,
    root_config: "BaseConfig",
    on_error: Literal["warn"] = "warn",
) -> _CkptCandidate | None: ...
@overload
def _load_ckpt_meta(
    path: Path,
    root_config: "BaseConfig",
    on_error: Literal["raise"],
) -> _CkptCandidate: ...
def _load_ckpt_meta(
    path: Path,
    root_config: "BaseConfig",
    on_error: Literal["warn", "raise"] = "warn",
):
    meta = CheckpointMetadata.from_file(path)
    if root_config.id != meta.run_id:
        error_msg = f"Skipping checkpoint {path} because it belongs to a different run"
        match on_error:
            case "warn":
                log.warn(error_msg)
            case "raise":
                raise ValueError(error_msg)
            case _:
                assert_never(on_error)
        return None
    return _CkptCandidate(meta, path)


def _checkpoint_candidates(
    root_config: "BaseConfig",
    trainer: LightningTrainer,
    *,
    include_hpc: bool = True,
):
    # Load the checkpoint directory, and throw if it doesn't exist.
    # This indicates a non-standard setup, and we don't want to guess
    # where the checkpoints are.
    ckpt_dir = root_config.directory.resolve_subdirectory(root_config.id, "checkpoint")
    if not ckpt_dir.is_dir():
        raise FileNotFoundError(
            f"Checkpoint directory {ckpt_dir} not found. "
            "Please ensure that the checkpoint directory exists."
        )

    # Load all checkpoints in the directory.
    # We can do this by looking for metadata files.
    for path in ckpt_dir.glob(f"*{METADATA_PATH_SUFFIX}"):
        if (meta := _load_ckpt_meta(path, root_config)) is not None:
            yield meta

    # If we have a pre-empted checkpoint, load it
    if include_hpc and (hpc_path := trainer._checkpoint_connector._hpc_resume_path):
        hpc_meta_path = Path(hpc_path).with_suffix(METADATA_PATH_SUFFIX)
        if (meta := _load_ckpt_meta(hpc_meta_path, root_config)) is not None:
            yield meta


def _additional_candidates(
    additional_candidates: Iterable[Path], root_config: "BaseConfig"
):
    for path in additional_candidates:
        if (
            meta := _load_ckpt_meta(path.with_suffix(METADATA_PATH_SUFFIX), root_config)
        ) is None:
            continue
        yield meta


def _resolve_checkpoint(
    config: CheckpointLoadingConfig,
    root_config: "BaseConfig",
    trainer: LightningTrainer,
):
    # We lazily load the checkpoint candidates to avoid loading them
    # if they are not needed.
    _ckpt_candidates: list[_CkptCandidate] | None = None

    def ckpt_candidates():
        nonlocal _ckpt_candidates, root_config, trainer

        if _ckpt_candidates is None:
            _ckpt_candidates = list(
                _checkpoint_candidates(
                    root_config, trainer, include_hpc=config.include_hpc
                )
            )
        return _ckpt_candidates

    # Iterate over the strategies and try to resolve the checkpoint.
    for strategy in config.strategies:
        match strategy:
            case UserProvidedPathCheckpointStrategyConfig():
                meta = _load_ckpt_meta(
                    strategy.path.with_suffix(METADATA_PATH_SUFFIX),
                    root_config,
                    on_error=strategy.on_error,
                )
                if meta is None:
                    continue
                return meta.ckpt_path
            case BestCheckpointStrategyConfig():
                candidates = [
                    *ckpt_candidates(),
                    *_additional_candidates(
                        strategy.additional_candidates, root_config
                    ),
                ]
                if not candidates:
                    log.warn(
                        "No checkpoint candidates found for `best` checkpoint strategy."
                    )
                    continue

                if (metric := strategy.metric or root_config.primary_metric) is None:
                    log.warn(
                        "No metric specified for `best` checkpoint strategy, "
                        "and no primary metric is set in the configuration. "
                        "Skipping strategy."
                    )
                    continue

                # Find the best checkpoint based on the metric.
                def metric_value(ckpt: _CkptCandidate):
                    assert metric is not None
                    if (
                        value := ckpt.meta.metrics.get(metric.validation_monitor)
                    ) is None:
                        raise ValueError(
                            f"Metric {metric.validation_monitor} not found in checkpoint metadata. "
                            f"Available metrics: {ckpt.meta.metrics.keys()}"
                        )
                    return value

                best_candidate = metric.best(candidates, key=metric_value)
                return best_candidate.ckpt_path
            case LastCheckpointStrategyConfig():
                candidates = [
                    *ckpt_candidates(),
                    *_additional_candidates(
                        strategy.additional_candidates, root_config
                    ),
                ]
                if not candidates:
                    log.warn(
                        "No checkpoint candidates found for `last` checkpoint strategy."
                    )
                    continue

                # Find the last checkpoint based on the criterion.
                def criterion_value(ckpt: _CkptCandidate):
                    match strategy.criterion:
                        case "global_step":
                            return ckpt.meta.global_step
                        case "runtime":
                            return ckpt.meta.training_time.total_seconds()
                        case _:
                            assert_never(strategy.criterion)

                last_candidate = max(candidates, key=criterion_value)
                return last_candidate.ckpt_path
            case _:
                assert_never(strategy)



from abc import ABC, abstractmethod
from collections import Counter
from collections.abc import Iterable
from dataclasses import dataclass
from typing import TYPE_CHECKING, TypeAlias, TypedDict

import nshconfig as C
from lightning.pytorch import Callback

if TYPE_CHECKING:
    from ..model.config import BaseConfig


class CallbackMetadataDict(TypedDict, total=False):
    ignore_if_exists: bool
    """If `True`, the callback will not be added if another callback with the same class already exists."""

    priority: int
    """Priority of the callback. Callbacks with higher priority will be loaded first."""


class CallbackMetadataConfig(C.Config):
    ignore_if_exists: bool = False
    """If `True`, the callback will not be added if another callback with the same class already exists."""

    priority: int = 0
    """Priority of the callback. Callbacks with higher priority will be loaded first."""


@dataclass(frozen=True)
class CallbackWithMetadata:
    callback: Callback
    metadata: CallbackMetadataConfig


ConstructedCallback: TypeAlias = Callback | CallbackWithMetadata


class CallbackConfigBase(C.Config, ABC):
    metadata: CallbackMetadataConfig = CallbackMetadataConfig()
    """Metadata for the callback."""

    def with_metadata(self, callback: Callback, **metadata: CallbackMetadataDict):
        return CallbackWithMetadata(
            callback=callback, metadata=self.metadata.model_copy(update=metadata)
        )

    @abstractmethod
    def create_callbacks(
        self, root_config: "BaseConfig"
    ) -> Iterable[Callback | CallbackWithMetadata]: ...


# region Config resolution helpers
def _create_callbacks_with_metadata(
    config: CallbackConfigBase, root_config: "BaseConfig"
) -> Iterable[CallbackWithMetadata]:
    for callback in config.create_callbacks(root_config):
        if isinstance(callback, CallbackWithMetadata):
            yield callback
            continue

        callback = config.with_metadata(callback)
        yield callback


def _filter_ignore_if_exists(callbacks: list[CallbackWithMetadata]):
    # First, let's do a pass over all callbacks to hold the count of each callback class
    callback_classes = Counter(callback.callback.__class__ for callback in callbacks)

    # Remove non-duplicates
    callbacks_filtered: list[CallbackWithMetadata] = []
    for callback in callbacks:
        # If `ignore_if_exists` is `True` and there is already a callback of the same class, skip this callback
        if (
            callback.metadata.ignore_if_exists
            and callback_classes[callback.callback.__class__] > 1
        ):
            continue

        callbacks_filtered.append(callback)

    return callbacks_filtered


def _process_and_filter_callbacks(
    callbacks: Iterable[CallbackWithMetadata],
) -> list[Callback]:
    callbacks = list(callbacks)

    # Sort by priority (higher priority first)
    callbacks.sort(key=lambda callback: callback.metadata.priority, reverse=True)

    # Process `ignore_if_exists`
    callbacks = _filter_ignore_if_exists(callbacks)

    return [callback.callback for callback in callbacks]


def resolve_all_callbacks(root_config: "BaseConfig"):
    callback_configs = [
        config
        for config in root_config._nshtrainer_all_callback_configs()
        if config is not None
    ]
    callbacks = _process_and_filter_callbacks(
        callback
        for callback_config in callback_configs
        for callback in _create_callbacks_with_metadata(callback_config, root_config)
    )
    return callbacks


# endregion
